# Copyright (c) 2016-2022 The Regents of the University of Michigan
# Part of GSD, released under the BSD 2-Clause License.

"""Define the current version of the gsd package.

This file is kept separate from __init__.py so that it can be imported and used
in other modules within the package.
"""

__version__ = "2.6.0"

__all__ = [
    '__version__',
]
